; -----------------------------------------------------------------------------
; Main routines to manage game execution (intro, menu, gameplay, finale).
; -----------------------------------------------------------------------------
; Coded by Einar Saukas
; -----------------------------------------------------------------------------

        org     41500

; -----------------------------------------------------------------------------
; Main routine that controls the entire game.
; -----------------------------------------------------------------------------
        call    intro_game              ; present intro
main_loop:
        ld      hl, chouyaku
        call    music_start             ; play music "Chouyaku"

        call    menu_game               ; display menu

        xor     a
        call    cls_scr

        call    music_stop              ; stop music

        ld      hl, stracker_ai+1084
        call    music_start             ; play music "Ai"

        call    play_game               ; play game, then stop music

        jr      main_loop               ; back to menu

; -----------------------------------------------------------------------------
; Set specified Soundtracker music address, then enable BIFROST* interrupt to
; play it automatically.
;
; Normally BIFROST* interrupt would be enabled/disabled in order to
; activate/deactivate multicolor. However this game works differently. Here,
; BIFROST* interrupt is used both to produce multicolor, and also to play AY
; music automatically in background. Therefore this game will enable/disable
; BIFROST* interrupt just to play/mute AY music, afterwards the enabled
; BIFROST* interrupt will be patched in order to activate/deactivate
; multicolor.
;
; Parameters:
;   HL: music address
; -----------------------------------------------------------------------------
music_start:
        ld      (stracker_ai+1), hl     ; select music
        halt
        call    stracker_ai             ; enable Soundtracker Player
        halt
        jp      64995                   ; enable BIFROST* interrupt to play Soundtracker

; -----------------------------------------------------------------------------
; Disable BIFROST* interrupt to stop playing Soundtracker music.
; -----------------------------------------------------------------------------
music_stop:
        halt
        call    65012                   ; disable BIFROST* interrupt to stop playing Soundtracker
        halt
        jp      stracker_ai             ; disable Soundtracker Player

; -----------------------------------------------------------------------------
; Play the game, then disable music afterwards.
;
; In "Practice" or "Players" mode, it plays a single match only (at previously
; selected level).
;
; In "Play!" mode, it plays a sequence of matches until player wins all levels
; (presenting the finale screen in this case), or until player is defeated.
;
; Uses:
;   play_mode: gameplay mode (0-2)
;   play_level: difficulty level (1-7)
;
; Destroys:
;   AF, BC, DE, HL, AF'
; -----------------------------------------------------------------------------
play_game:
        call    play_match              ; play selected match

        ld      a, (play_mode)
        and     a
        jr      nz, music_stop          ; stop unless "Play!" mode was selected

        ld      a, b                    ; score
        cp      128+96
        jr      nz, music_stop          ; stop if player lost match

        ld      hl, play_level
        inc     (hl)                    ; play_level++
        ld      a, 7
        cp      (hl)
        jr      nc, play_game           ; repeat until play_level > 7

        call    music_stop

        call    finale_print            ; print finale

; -----------------------------------------------------------------------------
; Wait for new key, i.e first wait until no key is pressed, then wait until any
; key is pressed.
;
; Destroys:
;   AF
; -----------------------------------------------------------------------------
wait_new_key:
        xor     a
        in      a, ($fe)
        cpl
        and     31
        jr      nz, wait_new_key

; -----------------------------------------------------------------------------
; Wait for new any key, i.e just wait until any key is pressed.
;
; Destroys:
;   AF
; -----------------------------------------------------------------------------
wait_any_key:
        xor     a
wait_any_key_loop:
        in      a, ($fe)
        cpl
        and     31
        jr      z, wait_any_key_loop
        ret

; -----------------------------------------------------------------------------
; Play a single game match.
;
; Returns:
;   B: 128+96 if player #1 won,
;      128-96 if player #2 won
;
; Uses:
;   play_level: difficulty level (1-7)
;   play_turn: current player turn (1-2)
;   play_computer: 2 if computer is player #2, 0 otherwise
;
; Destroys:
;   AF, BC, DE, HL, AF'
; -----------------------------------------------------------------------------
play_match:
        ; show in-game screen
        ld      hl, scr_backg
        ld      a, 0                    ; BORDER 0 (POKE HERE TO CHANGE IT!)
        call    dzx7_scr                ; display background screen

        ; update screen
        ld      a, (play_level)
        call    board_level_show        ; display level board sign

        ; initialize multicolor tiles
        ld      a, (play_level)
        call    play_tile_init

        ; initialize game
        call    mcolor_clear            ; initialize tilemap area
        call    mcolor_init             ; initialize multicolor area
        call    BRAINIAC_init           ; initialize BRAINIAC AI
        call    machine_init            ; initialize state machine

        ld      hl, 59021
        ld      ($fdfe), hl             ; enable BIFROST*

        call    play_start

        call    msg_result

        call    wait_new_key

        ld      hl, interrupt_handler
        ld      ($fdfe), hl             ; disable BIFROST*

        ret

; -----------------------------------------------------------------------------
;
; -----------------------------------------------------------------------------
play_start:
        ld      hl, play_turn
        ld      (hl), 2
play_loop:                              ; repeat {
        ld      hl, play_turn
        ld      a, (hl)
        xor     3
        ld      (hl), a                 ;     next player turn

        inc     hl
        cp      (hl)
        jr      nz, play_human          ;     if (turn == computer) {

        ld      hl, (play_level)
        call    BRAINIAC_best_move      ;         E = BRAINIAC_best_move

        ld      hl, machine_move2
        ld      (hl), e                 ;         store move E

        ld      a, 1
        call    play_wait_anim_finished ;         ensure opponent's previous animation was finished

        jr      play_endif              ;     } else {
play_human:

        xor     3
        call    play_wait_anim_finished ;         ensure opponent's previous animation was finished

        xor     a
        ld      (23560), a              ;         reset last pressed key
play_wait_new_move:
        ld      a, (23560)
        sub     '1'
        jr      c, play_wait_new_move
        cp      4
        jr      nc, play_wait_new_move  ;         repeast until player chooses valid move E

        ld      e, a
        ld      a, (play_turn)
        call    play_wait_anim_finished
        ld      (hl), e                 ;         store move E

        ld      a, e
        call    board_move_show         ;         display move immediately

play_endif:                             ;     }

        ld      a, (play_turn)
        ld      c, a
        ld      a, e
        call    BRAINIAC_play           ;     BRAINIAC_play(E, turn)

        call    BRAINIAC_check_board    ;     boardScore = BRAINIAC_check_board

        cp      128+96
        jr      z, play_gameover
        cp      128-96
        jr      nz, play_loop           ; } until (boardScore == 128+96 or boardScore == 128-96)
play_gameover:

        ld      b, a                    ; preserve boardScore

        ld      a, (play_turn)
        xor     3
        call    play_wait_anim_finished
        ld      (hl), $ff               ; set opponent's move to $ff to indicate "game over"

        ld      a, (play_turn)          ; wait until current player animation was finished, then return boardScore
play_wait_anim_finished:
        add     a, (machine_move1-1)%256
        ld      l, a
        ld      h, (machine_move1-1)/256
play_wait_loop:
        bit     7, (hl)
        jr      z, play_wait_loop
        ret

; -----------------------------------------------------------------------------
; Decompress multicolor tileset corresponding to specified level.
;
; Parameters:
;   A: difficulty level (1-7)
;
; Destroys:
;   AF, BC, DE, HL, AF'
; -----------------------------------------------------------------------------
play_tile_init:
        add     a, a
        ld      c, a
        ld      b, 0
        ld      hl, play_ctiles-2
        add     hl, bc
        ld      e, (hl)
        inc     hl
        ld      d, (hl)
        ex      de, hl
        ld      de, TILE_IMAGES
        jp      dzx7_agilercs           ; decompress multicolor tiles

; -----------------------------------------------------------------------------
; Table of compressed BIFROST* tiles address for each level
; -----------------------------------------------------------------------------
play_ctiles:
        dw      ctile_lv1
        dw      ctile_lv2
        dw      ctile_lv3
        dw      ctile_lv4
        dw      ctile_lv5
        dw      ctile_lv6
        dw      ctile_lv7

; -----------------------------------------------------------------------------
; Internal variables
; -----------------------------------------------------------------------------
play_mode:
        db      0                       ; gameplay mode (from main menu)
play_level:
        db      0                       ; difficulty level (from sub-menu)
play_turn:
        db      2                       ; current player turn (1-2)
play_computer
        db      0                       ; 2 if computer is player #2, 0 otherwise

; -----------------------------------------------------------------------------
