; -----------------------------------------------------------------------------
; Customized BIFROST* Engine. Parts of the original BIFROST*, not used in this 
; game, are replaced here with specialized routines that allow drawing a 
; sequence of multicolor tiles outside the regular multicolor area.
; -----------------------------------------------------------------------------
; Coded by Einar Saukas
; -----------------------------------------------------------------------------

; -----------------------------------------------------------------------------
; Compile BIFROST* version v1.2/H, but reconfigured to use different values for
; TILE_IMAGES and SHIFT_COLUMNS
; -----------------------------------------------------------------------------
TILE_IMAGES     EQU 53013
SHIFT_COLUMNS   EQU 4

INCLUDE "BIFROST_H.asm"

; -----------------------------------------------------------------------------
; Replace BIFROST* routine "draw_at_last_col" since it's never used in this
; game.
        org     $df81
; -----------------------------------------------------------------------------

; -----------------------------------------------------------------------------
; Disable the custom multicolor render, restoring BIFROST* to its normal
; behavior. Caller must disable interrupts or synchronize with them, to ensure
; this routine won't be interrupted during execution.
;
; Destroys:
;   AF, HL
; -----------------------------------------------------------------------------
mcolor_disable:
        ld      a, $10
        ld      (wait_raster), a
        ld      hl, $edfe
        ld      (wait_raster+1), hl
        ret

; -----------------------------------------------------------------------------
; Replace first 16 iterations of the original renderer. Instead of drawing the
; first row of 9 multicolor tiles between columns 1 to 18, this replacement
; renderer can draw 3 multicolor tiles starting at any column on first row.
;
; Destroys:
;   AF, BC, DE, HL, AF', BC', DE', HL'
; -----------------------------------------------------------------------------
mcolor_render:
        ld      (exit_raster+1), sp
mcolor_render_tile1:
        ld      hl, 0                   ; tile1*64
        ld      bc, TILE_IMAGES+32
        exx
        sbc     hl, hl                  ; extra delay

; update multicolor attributes of 3 tiles during 1st row rendering (8 pixel lines)
        ld      b, 8
mcolor_render_row1:
        add     hl, hl                  ; extra delay
        exx
        add     hl, bc
        ld      a, (hl)
        ex      af, af'
        inc     hl
        ld      a, (hl)
        ex      af, af'
mcolor_render_delta21r1:
        ld      bc, 0                   ; from 2nd byte frame1 to 1st byte frame2 = (tile2-tile1)*64-1
        add     hl, bc
mcolor_render_delta32r1:
        ld      bc, 0                   ; (from 3rd byte frame2 to 1st byte frame3) +2 = (tile3-tile2)*64
mcolor_render_scr1:
        ld      de, $5820               ; screen address of 1st tile in row1
mcolor_render_instr1r1:
        ld      (de), a                 ; or "LD A,(DE)" if there's no left side ($12 or $1a)
        inc     e
        ex      af, af'
        ld      (de), a
        inc     e
        ldi
        ldi
        add     hl, bc
        ldi
        ld      a, (hl)
mcolor_render_instr3r1:
        ld      (de), a                 ; or "LD A,(DE)" if there's no right side ($12 or $1a)
mcolor_render_delta31r1:
        ld      bc, 0                   ; from 2nd byte frame3 to 1st byte frame1 +2 = (tile1-tile3)*64+1
        exx
        djnz    mcolor_render_row1

; update multicolor attributes of 3 tiles during 2nd row rendering (8 pixel lines)
        ld      b, 8
mcolor_render_row2:
        add     hl, hl                  ; extra delay
        exx
        add     hl, bc
        ld      a, (hl)
        ex      af, af'
        inc     hl
        ld      a, (hl)
        ex      af, af'
mcolor_render_delta21r2:
        ld      bc, 0                   ; from 2nd byte frame1 to 1st byte frame2 = (tile2-tile1)*64-1
        add     hl, bc
mcolor_render_delta32r2:
        ld      bc, 0                   ; (from 3rd byte frame2 to 1st byte frame3) +2 = (tile3-tile2)*64
mcolor_render_scr2:
        ld      de, $5840               ; screen address of 1st tile in row2
mcolor_render_instr1r2:
        ld      (de), a                 ; or "LD A,(DE)" if there's no left side ($12 or $1a)
        inc     e
        ex      af, af'
        ld      (de), a
        inc     e
        ldi
        ldi
        add     hl, bc
        ldi
        ld      a, (hl)
mcolor_render_instr3r2:
        ld      (de), a                 ; or "LD A,(DE)" if there's no right side ($12 or $1a)
mcolor_render_delta31r2:
        ld      bc, 0                   ; from 2nd byte frame3 to 1st byte frame1 +2 = (tile1-tile3)*64+1
        exx
        djnz    mcolor_render_row2

        jp      race_raster+(16*41)

; -----------------------------------------------------------------------------
; State machine variables
; -----------------------------------------------------------------------------
machine_delay:
        db      0                       ; countdown for next machine state (in frames)
machine_frame:
        db      0                       ; current tile animation frame
machine_column:
        db      0                       ; current tile column position
machine_dir:
        db      0                       ; current movement direction (2 or -2)
machine_move1:
        db      0                       ; last player #1 selected column
machine_move2:
        db      0                       ; last player #2 selected column
machine_move:
        dw      0                       ; address of last selected column for current player
machine_buffer:
        db      0, 0, 0, 0, 0, 0, 0, 0  ; copy of current player's selected column

; -----------------------------------------------------------------------------
; Erase remaining unused area
; -----------------------------------------------------------------------------
REPT 3
        nop
ENDM

; -----------------------------------------------------------------------------
; Replace BIFROST* routine "draw_at_even_col_dec" since it's never used in this
; game.
        org     $e2f3
; -----------------------------------------------------------------------------

; -----------------------------------------------------------------------------
; Draws a sequence of 3 tiles at any arbitrary column in the top-most row. The
; special multicolor render is automatically activated whenever this method is
; called.
;
; Params:
;   B: tile1 (first tile in the sequence)
;   E: column (-5 to 29)
;
; Destroys:
;   AF, BC, DE, HL
; -----------------------------------------------------------------------------
mcolor_draw:
        ld      a, $c3                  ; instruction "JP"
        ld      (wait_raster), a
        ld      hl, mcolor_render
        ld      (wait_raster+1), hl     ; intercept beginning of BIFROST* multicolor, inserting a "JP mcolor_render"

        ld      a, $12                  ; set instruction "LD (DE),A" in custom render to display all tile attributes
        ld      (mcolor_render_instr1r1), a
        ld      (mcolor_render_instr1r2), a
        ld      (mcolor_render_instr3r1), a
        ld      (mcolor_render_instr3r2), a

        ld      a, b
        cp      32
        jr      nc, mcolor_draw_walking ; if (tile1 isn't walking animation) {
        add     a, 4
        ld      c, a                    ;     tile2 = tile1 + 4;
        ld      d, 41                   ;     tile3 = empty;
        jr      mcolor_draw_attr
mcolor_draw_walking:                    ; } else {
        inc     a
        ld      c, a                    ;     tile2 = tile1 + 1;
        inc     a
        ld      d, a                    ;     tile3 = tile2 + 1;
        cp      55
        jr      nz, mcolor_draw_not55   ;     if (tile3 == 55)
        ld      d, 41                   ;         tile3 = empty;
mcolor_draw_not55:
        ld      a, e
        cp      -5                      ;     if (col < 0) {
        jr      c, mcolor_draw_notleft  ;         // tiles are partially hidden on left side
        cp      -1
        jr      z, mcolor_draw_1stleft  ;         if (col == -3 || col == -5) {
        ld      b, c                    ;             tile1 = tile2;
        ld      c, d                    ;             tile2 = tile3;
        ld      d, 41                   ;             tile3 = empty;
        cp      -5
        jr      nz, mcolor_draw_2ndleft ;             if (col == -5) {
        ld      b, c                    ;                 tile1 = tile2;
        ld      c, d                    ;                 tile2 = empty;
mcolor_draw_2ndleft:                    ;             }
        ld      e, -1                   ;             col = -1;
mcolor_draw_1stleft:                    ;         }
        ld      a, $1a                  ;         set first instruction "LD A,(DE)" to display only half tile1 attributes
        ld      (mcolor_render_instr1r1), a
        ld      (mcolor_render_instr1r2), a
        jr      mcolor_draw_attr
mcolor_draw_notleft:
        cp      25                      ;     } else if (col >= 25) {
        jr      c, mcolor_draw_attr     ;         // tiles are partially hidden on right side
        jr      z, mcolor_draw_1stright ;         if (col == 27 || col == 29) {
        ld      d, c                    ;             tile3 = tile2;
        ld      c, b                    ;             tile2 = tile1;
        ld      b, 41                   ;             tile1 = empty;
        cp      29
        jr      nz, mcolor_draw_2ndright;             if (col == 29) {
        ld      d, c                    ;                 tile3 = tile2;
        ld      c, b                    ;                 tile2 = empty;
mcolor_draw_2ndright:                   ;             }
        ld      e, 25                   ;             col = 25;
mcolor_draw_1stright:                   ;         }
        ld      a, $1a                  ;         set last instruction "LD A,(DE)" to display only half tile3 attributes
        ld      (mcolor_render_instr3r1), a
        ld      (mcolor_render_instr3r2), a
mcolor_draw_attr:                       ;     }
        push    de                      ; }
        push    bc

        ld      a, e                    ; update custom render with initial screen position for tiles
        add     a, $20
        ld      (mcolor_render_scr1+1), a
        add     a, $20
        ld      (mcolor_render_scr2+1), a

        ld      h, b                    ; H = tile1
        xor     a
        srl     h
        rra
        srl     h
        rra
        ld      l, a                    ; HL = tile1*64
        ld      (mcolor_render_tile1+1), hl

        ld      a, d                    ; A = tile3
        sub     c                       ; A = tile3 - tile2
        ld      d, a
        xor     a
        sra     d
        rra
        sra     d
        rra
        ld      e, a                    ; DE = (tile3 - tile2)*64
        ld      (mcolor_render_delta32r1+1), de
        ld      (mcolor_render_delta32r2+1), de

        ld      a, c                    ; A = tile2
        sub     b                       ; A = tile2 - tile1
        ld      h, a
        xor     a
        sra     h
        rra
        sra     h
        rra
        ld      l, a
        dec     hl                      ; HL = (tile2 - tile1)*64 - 1
        ld      (mcolor_render_delta21r1+1), hl
        ld      (mcolor_render_delta21r2+1), hl

        add     hl, de                  ; HL = (tile3 - tile1)*64 - 1
        xor     a
        sub     l
        ld      l, a
        sbc     a, a
        sub     h
        ld      h, a                    ; HL = (tile1 - tile3)*64 + 1
        ld      (mcolor_render_delta31r1+1), hl
        ld      (mcolor_render_delta31r2+1), hl

        pop     bc
        pop     de
        push    de
        push    bc
        ld      d, b
        call    mcolor_draw_bmp         ; draw tile1 bitmap

        pop     bc
        pop     de
        inc     e
        inc     e
        push    de
        ld      d, c
        call    mcolor_draw_bmp         ; draw tile2 bitmap

        pop     de
        inc     e
        inc     e                       ; draw tile3 bitmap

; -----------------------------------------------------------------------------
; Draws the bitmap part of a single tile at any arbitrary column in the top-most
; row.
;
; Params:
;   D = tile
;   E = column (-1 to 25)
; -----------------------------------------------------------------------------
mcolor_draw_bmp:
; calculate screen bitmap lookup address
        ld      a, e
        sub     4
        ld      (mcolor_draw_bmp_col+1), a ; set column

        ld      hl, $a0ed               ; instruction "LDI"
        cp      -5
        jr      nz, mcolor_draw_bmp_left
        ld      hl, $1323               ; instructions "INC HL / INC DE"
        
; draw left side bitmap of specified tile
mcolor_draw_bmp_left:
        ld      (mcolor_draw_bmp_adjust_left), hl
        cp      25
        ld      a, $a0                  ; instruction "LDI"
        jr      nz, mcolor_draw_bmp_right
        inc     a                       ; instruction "CPI"

; draw right side bitmap of specified tile
mcolor_draw_bmp_right:
        ld      (mcolor_draw_bmp_adjust_right+1), a

; calculate tile image address
        xor     a
        srl     d
        rra
        srl     d
        rra
        ld      e, a                    ; DE = tile*64
        ld      hl, TILE_IMAGES
        add     hl, de                  ; HL = 48500+tile*64

        ld      bc, $10ff

        ld      (mcolor_draw_bmp_exit+1), sp
        ld      sp, bitmaps+32

; draw bitmap lines
mcolor_draw_bmp_next_line:
        pop     de
        ld      a, e
mcolor_draw_bmp_col:
        add     a, 0
        ld      e, a
mcolor_draw_bmp_adjust_left:
        ldi                             ; or "INC HL" / "INC DE"
mcolor_draw_bmp_adjust_right:
        ldi                             ; or "CPI"
        djnz    mcolor_draw_bmp_next_line
mcolor_draw_bmp_exit:
        ld      sp, 0
        ret

; -----------------------------------------------------------------------------
; Erase remaining unused area
; -----------------------------------------------------------------------------
REPT 4
        nop
ENDM


; -----------------------------------------------------------------------------
; Replace BIFROST* routine "draw_at_even_col_inc" since it's never used in this
; game.
        org     $e414
; -----------------------------------------------------------------------------

; -----------------------------------------------------------------------------
; Set all tilemap positions to BIFROST_DISABLED
;
; Destroys:
;   HL
; -----------------------------------------------------------------------------
mcolor_clear:
        ld      hl, $ff51
mcolor_clear_next:
        ld      (hl), h
        dec     l
        jr      nz, mcolor_clear_next
        ret

; -----------------------------------------------------------------------------
; Initialize multicolor area for a new game match. Most multicolor positions
; are painted black, except for horizontal platform tiles (initialized at
; different frames) and left side column of tiles (painted with same attributes
; from regular background screen).
;
; Destroys:
;   AF, BC, DE, HL
; -----------------------------------------------------------------------------
mcolor_init:
        ; set horizontal platform tiles
        ld      hl, $ff0b
        ld      b, 8
mcolor_init_loop:
        ld      a, r
        and     3
        or      60                      ; pseudo-random tile between 60-63
        ld      (hl), a
        inc     hl
        djnz    mcolor_init_loop

        ; sync with raster beam to avoid crashes
        halt

        ; left side column not used, so paint attributes from regular background screen
        ld      hl, 22626+SHIFT_COLUMNS ; first attribute address AT 3,6
        ld      de, 32*256+1            ; D=lin, E=col
mcolor_init_left_column:
        push    hl
        push    de
        ld      c, (hl)
        call    fill_tile_attr          ; paint multicolor block (C=attr, D=lin, E=col)
        pop     de
        pop     hl
        push    hl
        push    de
        dec     e
        dec     l
        ld      c, (hl)
        call    fill_tile_attr          ; paint multicolor block (C=attr, D=lin, E=col)
        pop     de
        pop     hl
        ld      bc, 32
        add     hl, bc
        ld      a, d
        add     a, 8
        ld      d, a
        cp      $98
        jr      nz, mcolor_init_left_column

        ; draw horizontal platform
        ld      de, 32*256+1
mcolor_init_platform:
        inc     e
        inc     e
        push    de
        ld      a, 60
        call    draw_tile
        pop     de
        bit     4, e
        jr      z, mcolor_init_platform

        ; sync with raster beam to avoid crashes
        halt

        ; paint remaining tile positions as blank
mcolor_init_next_row:
        ld      hl, 16*256-16
        add     hl, de
        ex      de, hl
mcolor_init_next_col:
        inc     e
        inc     e
        push    de
        ld      c, 0
        call    fill_tile_attr          ; paint multicolor block (C=attr, D=lin, E=col)
        pop     de
        bit     4, e
        jr      z, mcolor_init_next_col
        ld      a, d
        cp      144
        jr      nz, mcolor_init_next_row
        ret

; -----------------------------------------------------------------------------
; Check entire board for aligned avatars. Whenever finding 4 (or more) aligned
; avatars from same player, highlight those avatars by displaying a distinct
; special idle animation.
;
; Destroys:
;   AF, BC, DE, HL, AF'
; -----------------------------------------------------------------------------
mcolor_aligned:
        ; reset all idle animations
        ld      hl, 65300
        ld      a, $1f
        ld      b, 62
mcolor_aligned_reset_next:
        cp      (hl)
        jr      c, mcolor_aligned_not_idle
        res     4, (hl)
mcolor_aligned_not_idle:
        inc     l
        djnz    mcolor_aligned_reset_next

        ; process horizontal alignments
        ld      de, 2                   ; horizontal towards left
        ld      a, 17                   ; for each last 7 rows {
mcolor_aligned_next:
        add     a, 9
        call    mcolor_highlight        ;     process row
        ex      af, af'
        cp      80
        jr      nz, mcolor_aligned_next ; }

        ; process vertical and diagonal alignments
mcolor_aligned_prev:                    ; for each last 4 rows in descending order {
        ld      e, 11                   ;     diagonal towards up-left
        call    mcolor_highlight        ;     process diagonal
        ex      af, af'
        ld      e, 9                    ;     vertical towards up
        call    mcolor_highlight        ;     process 4th pair of columns
        ex      af, af'
        sub     2
        call    mcolor_highlight        ;     process 3rd pair of columns
        ex      af, af'
        sub     2
        call    mcolor_highlight        ;     process 2nd pair of columns
        ex      af, af'
        sub     2
        call    mcolor_highlight        ;     process 1st pair of columns
        ex      af, af'
        ld      e, 7                    ;     diagonal towards up-right
        call    mcolor_highlight        ;     process diagonal
        ex      af, af'
        sub     3                       ;     previous row
        cp      44
        jr      nz, mcolor_aligned_prev ; }
        ret

; -----------------------------------------------------------------------------
; Search for 4 identical avatars in specified direction, starting at specified
; initial tile position. If found, then update these avatars to display a
; distinct special idle animation.
;
; Parameters:
;   A: specified initial tile map position (lower byte address)
;   DE: specified direction (to be subtracted from position)
;   H: must be 255
;
; Returns:
;   A': specified initial tile map position (lower byte address)
;
; Destroys:
;   F, BC, L, F'
; -----------------------------------------------------------------------------
mcolor_highlight:
        ld      l, a
        ex      af, af'
        ld      a, (hl)
        cp      $20
        ret     nc
        and     $e8
        ld      c, a
        ld      b, 3
mcolor_highlight_check:
        sbc     hl, de
        ld      a, (hl)
        and     $e8
        cp      c
        ret     nz
        djnz    mcolor_highlight_check
        ld      b, 4
mcolor_highlight_update:
        set     4, (hl)
        inc     l
        set     4, (hl)
        dec     l
        add     hl, de
        djnz    mcolor_highlight_update
        ret

; -----------------------------------------------------------------------------
; Replacement interrupt handler to play AY music and execute state machine.
; It's called directly from interrupts (when BIFROST* is disabled) or called
; immediately after BIFROST* renderer (when BIFROST* is enabled).
; -----------------------------------------------------------------------------
interrupt_handler:
        push    af                      ; preserve registers
        ex      af, af'
        push    af
        push    bc
        push    de
        push    hl
        push    ix

        call    44707+6                 ; play next tick from Soundtracker Player

state_machine:
        ld      hl, machine_delay
        dec     (hl)                    ; countdown for next machine state
state_caller:
        call    z, interrupt_handler-1  ; invoke next machine state

        pop     ix                      ; restore registers
        pop     hl
        pop     de
        pop     bc
        pop     af
        ex      af, af'
        pop     af
        jp      $38                     ; exit through RST $38

; -----------------------------------------------------------------------------
; Erase remaining unused area
; -----------------------------------------------------------------------------
REPT 7
        nop
ENDM

; -----------------------------------------------------------------------------
