; -----------------------------------------------------------------------------
; Display and control selection menus.
; -----------------------------------------------------------------------------
; Coded by Einar Saukas
; -----------------------------------------------------------------------------

; -----------------------------------------------------------------------------
; Display and control main menu screen where player chooses gameplay mode
; "Play!", "Practice", or "Players". In last 2 cases, also display and control
; sub-menu screen where player also chooses difficulty level.
;
; Uses:
;   play_mode: gameplay mode (0-2)
;   play_level: difficulty level (1-7)
;   play_computer: 2 if computer is player #2, 0 otherwise
;
; Destroys:
;   AF, BC, DE, HL, AF'
; -----------------------------------------------------------------------------
menu_game:
        ; show main menu screen
        ld      hl, scr_menu1
        ld      a, 7                    ; BORDER 7
        call    dzx7_scr                ; display main menu screen

        ; initialize multicolor tiles
        ld      hl, ctile_menu
        ld      de, TILE_IMAGES
        call    dzx7_agilercs           ; decompress main menu multicolor tiles

        ; initialize multicolor attributes
        call    menu_screen_init

        ; build menu content
        call    menu1_build

        ; show menu content ASAP
        call    menu_show_tiles

        ; enable multicolor
        ld      hl, 59021
        ld      ($fdfe), hl             ; enable BIFROST*

        ; process main menu
        call    menu1_input

        ; disable multicolor (but keep music playing)
        ld      hl, interrupt_handler   ; interrupt_handler
        ld      ($fdfe), hl             ; configure BIFROST* start to call interrupt_handler directly

        ; store selected menu option
        ld      hl, play_mode
        ld      (hl), c                 ; play_mode
        inc     hl
        ld      (hl), 1                 ; play_level
        ld      a, c
        cpl
        and     $02
        ld      (play_computer), a      ; play_computer

        ; skip sub-menu if "Play!" mode was selected
        ld      a, c
        and     a
        ret     z

        ; show sub-menu screen
        ld      hl, scr_menu2
        ld      a, 6                    ; BORDER 6
        call    dzx7_scr                ; display sub-menu screen

        ; initialize multicolor attributes
        call    menu_screen_init

        ; initialize multicolor tiles
        ld      hl, ctile_sample
        ld      de, TILE_IMAGES
        call    dzx7_agilercs           ; decompress sub-menu multicolor tiles

        ; build menu content
        call    menu2_build

        ; show menu content ASAP
        call    menu_show_tiles

        ; set all tiles to BIFROST_DISABLED
        call    mcolor_clear

        ; enable multicolor
        ld      hl, 59021
        ld      ($fdfe), hl             ; enable BIFROST*

        ; process main menu
        call    menu2_input

        ; disable multicolor (but keep music playing)
        ld      hl, interrupt_handler   ; interrupt_handler
        ld      ($fdfe), hl             ; configure BIFROST* start to call interrupt_handler directly

        ; store selected level option
        ld      a, c
        ld      (play_level), a

        ; done!
        ret

; -----------------------------------------------------------------------------
; Put static tiles for main menu in multicolor area on screen.
;
; Destroys:
;   HL
; -----------------------------------------------------------------------------
menu1_build:
        ld      hl, $80a0               ; "Play!" tile (left side)
        ld      ($ff1d), hl
        ld      hl, $a084               ; "Play!" tile (right side)
        ld      ($ff1f), hl
        ld      hl, $88a0               ; "Practice" tile (left side)
        ld      ($ff2f), hl
        ld      hl, $a08c               ; "Practice" tile (right side)
        ld      ($ff31), hl
        ld      hl, $9490               ; "Players" tile (left side)
        ld      ($ff41), hl
        ld      hl, $9c98               ; "Players" tile (right side)
        ld      ($ff43), hl
        ret

; -----------------------------------------------------------------------------
; Control main menu and obtain selected option.
;
; Returns:
;   C: selected main menu option (0-2)
;
; Destroys:
;   AF, B, DE, HL, AF', BC', DE', HL'
; -----------------------------------------------------------------------------
menu1_input:
        ; start at first menu option
        ld      c, 0                    ; currently selected option (0-2)
        ld      hl, $ff1d               ; tilemap address of currently selected tiles
        ld      de, $400b               ; selected option position in multicolor area

        ; animate sequence of 4 tiles in current menu option
menu1_next:
        ld      b, 4
menu1_anim:
        res     7, (hl)                 ; animate this tile
        inc     l
        djnz    menu1_anim

        ; paint current menu option name as white
        halt
        ld      b, 4
menu1_white:
        push    de
        exx
        pop     de
        ld      c, 7                    ; INK 7
        call    fill_tile_attr          ; paint multicolor block (C=attr, D=lin, E=col)
        exx
        inc     e
        inc     e
        djnz    menu1_white

        ; ignore previously pressed keys
menu1_wait_nokey:
        xor     a
        in      a, ($fe)
        cpl
        and     31
        jr      nz, menu1_wait_nokey

        ; wait for ENTER or SPACE
        xor     a
        ld      (23560), a              ; reset last pressed key
menu1_wait_key:
        ld      a, (23560)
        cp      13
        ret     z                       ; if ENTER then exit
        cp      32
        jr      nz, menu1_wait_key      ; repeat until player chooses ENTER or SPACE

        ; paint current menu option name as dark blue
        halt
        ld      b, 4
menu1_blue:
        dec     e
        dec     e
        push    de
        exx
        pop     de
        ld      c, 1                    ; INK 1
        call    fill_tile_attr          ; paint multicolor block (C=attr, D=lin, E=col)
        exx
        djnz    menu1_blue

        ; stop animating sequence of 4 tiles in current menu option
        ld      b, 4
menu1_static:
        set     7, (hl)                 ; stop animating this tile
        dec     l
        djnz    menu1_static

        ; move to next menu option
        inc     c
        ld      a, d
        add     a, 32
        ld      d, a
        ld      a, l
        add     a, 18
        ld      l, a
        cp      $53
        jr      nz, menu1_next
        jr      menu1_input

; -----------------------------------------------------------------------------
; Put static tiles for sub-menu in multicolor area on screen (4 multicolor
; tiles per pair of char rows, located at screen columns 7+4, 9+4, 13+4, 15+4
; respectively).
;
; Destroys:
;   AF, HL
; -----------------------------------------------------------------------------
menu2_build:
        ld      hl, $ff0d
        ld      a, $80
menu2_build_next:
        ld      (hl), a
        inc     l
        inc     a
        ld      (hl), a
        inc     l
        inc     l
        inc     a
        ld      (hl), a
        inc     l
        inc     a
        ld      (hl), a
        inc     l
        inc     l
        inc     l
        inc     l
        inc     l
        inc     a
        cp      $9c
        jr      nz, menu2_build_next
        ret

; -----------------------------------------------------------------------------
; Control sub-menu and obtain selected option.
;
; Notice that the multicolor area is shifted 4 columns to the right, occupying
; screen columns 1+4 to 18+4. The sub-menu text is even larger, occupying
; screen columns 4 to 27. Therefore both regular screen areas and multicolor
; areas must be painted, depending on currently selected option.
;
; The 4 multicolor tiles in each sub-menu option are located at columns 7+4,
; 9+4, 13+4, and 15+4.
;
; Returns:
;   C: selected sub-menu option (1-7)
;
; Destroys:
;   AF, B, DE, HL, AF', BC', DE', HL'
; -----------------------------------------------------------------------------
menu2_input:
        ; start at first menu option
        ld      c, 1                    ; currently selected option (1-7)
        ld      hl, $ff0d               ; tilemap address of currently selected tiles
        ld      d, 32                   ; selected option position in multicolor area

        ; paint current menu option name as white
menu2_next:
        halt
        ld      e, 1
menu2_white:
        push    de
        exx
        pop     de
        ld      c, 7                    ; INK 7
        call    fill_tile_attr          ; paint multicolor block (C=attr, D=lin, E=col)
        exx
menu2_white_skip:
        inc     e
        inc     e

        ; paint all multicolor columns except 7+4, 9+4, 13+4 and 15+4
        ld      a, e
        cp      7
        jr      c, menu2_white
        cp      11
        jr      z, menu2_white
        cp      17
        jr      z, menu2_white
        jr      c, menu2_white_skip

        ; paint also parts of current menu option name at regular screen as white
        ld      b, 7                    ; INK 7
        call    menu2_side_attr

        ; decompress multicolor tiles corresponding to current option
        ld      a, c
        exx
        call    play_tile_init          ; decompress multicolor tiles
        exx

        ; put 4 animated tiles in current menu option
        ld      (hl), 0
        inc     l
        ld      (hl), 4
        inc     l
        inc     l
        ld      (hl), 8
        inc     l
        ld      (hl), 12

        ; wait for ENTER or SPACE
        xor     a
        ld      (23560), a              ; reset last pressed key
menu2_wait_key:
        ld      a, (23560)
        cp      13
        ret     z                       ; if ENTER then exit
        cp      32
        jr      nz, menu2_wait_key      ; repeat until player chooses ENTER or SPACE

        ; paint current menu option name as dark blue
        halt
        dec     e
        dec     e
menu2_blue:
        push    de
        exx
        pop     de
        ld      c, 1                    ; INK 1
        call    fill_tile_attr          ; paint multicolor block (C=attr, D=lin, E=col)
        exx
menu2_blue_skip:
        ld      (hl), $ff               ; stop animating this tile
        dec     l
        dec     e
        jr      z, menu2_blue_end
        dec     e

        ; paint all multicolor columns except 7+4, 9+4, 13+4 and 15+4
        ld      a, e
        cp      7
        jr      c, menu2_blue
        cp      11
        jr      z, menu2_blue
        jr      menu2_blue_skip

menu2_blue_end:
        ; paint also parts of current menu option name at regular screen as dark blue
        ld      b, 1                    ; INK 1
        call    menu2_side_attr

        ; move to next menu option
        inc     c
        ld      a, l
        add     a, 14
        ld      l, a
        ld      a, d
        add     a, 16
        ld      d, a
        cp      144
        jr      z, menu2_input
        jr      menu2_next

; -----------------------------------------------------------------------------
; Certain sub-menu option names are too large, so some parts are located in
; regular screen outside multicolor area. This routine updates regular screen 
; attributes for current sub-menu option accordingly, so the player won't 
; distinguish between multicolor and regular screen areas.
;
; Parameters:
;   B: specified attribute
;   D: pixel line for current sub-menu option
;
; Destroys:
;   AF
; -----------------------------------------------------------------------------
menu2_side_attr:
        push    hl
        ld      h, $16
        ld      l, d                    ; HL = $1600 + lin
        inc     l                       ; HL = $1601 + lin
        add     hl, hl                  ; HL = $2c02 + 2*lin
        add     hl, hl                  ; HL = $5804 + 4*lin
menu2_side_loop:
        ld      (hl), b                 ; update column 4 with specified attribute
        ld      a, l
        add     a, 19
        ld      l, a
        ld      (hl), b                 ; update column 23 with specified attribute
        inc     l
        ld      (hl), b                 ; update column 24 with specified attribute
        inc     l
        ld      (hl), b                 ; update column 25 with specified attribute
        inc     l
        ld      (hl), b                 ; update column 26 with specified attribute
        inc     l
        ld      (hl), b                 ; update column 27 with specified attribute
        sub     51
        ld      l, a                    ; move HL to column 4 at previous char row
        jr      nc, menu2_side_skip
        dec     h                       ; necessary adjust at "screen third" boundary
menu2_side_skip:
        and     $20
        jr      nz, menu2_side_loop     ; repeat twice since each tile occupies 2 rows
        pop     hl
        ret

; -----------------------------------------------------------------------------
; Update all 9x9 multicolor tile positions on screen ASAP, so all multicolor 
; tiles will appear at once when multicolor is activated.
;
; Destroys:
;   AF, BC, DE, HL
; -----------------------------------------------------------------------------
menu_show_tiles:
        ld      c, 4
menu_show_frame:
        halt
        ld      b, 7
menu_show_loop:
        push    bc
        call    show_next3
        pop     bc
        djnz    menu_show_loop
        dec     c
        jr      nz, menu_show_frame
        ret

; -----------------------------------------------------------------------------
; Initialize multicolor area to replicate same attributes from regular screen.
;
; Destroys:
;   AF, BC, DE, HL
; -----------------------------------------------------------------------------
menu_screen_init:
        ; set all tiles to BIFROST_DISABLED
        call    mcolor_clear

        ; copy regular screen attributes to multicolor
        ld      hl, 22561+SHIFT_COLUMNS ; first attribute address
        ld      d, 16                   ; D=lin
menu_screen_next_row:
        ld      e, 1                    ; E=col
        bit     3, d
        jr      nz, menu_screen_next_col
        halt
menu_screen_next_col:
        push    hl
        push    de
        ld      c, (hl)                 ; C=attr (from regular screen)
        call    fill_tile_attr          ; paint multicolor block (C=attr, D=lin, E=col)
        pop     de
        pop     hl
        inc     l
        inc     e
        ld      a, e
        cp      19
        jr      nz, menu_screen_next_col
        ld      bc, 14
        add     hl, bc
        ld      a, d
        add     a, 8
        ld      d, a
        cp      160
        jr      nz, menu_screen_next_row
        ret

; -----------------------------------------------------------------------------
